# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '05 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

import sys
import logging
import hashlib
import argparse
from elasticsearch import Elasticsearch

from cci_os_worker import logstream
from .utils import load_config

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

def get_args():

    parser = argparse.ArgumentParser(description='Entrypoint for the CCI OS Worker on the CMD Line')
    parser.add_argument('conf', type=str, help='Path to Yaml config file for Elasticsearch')

    parser.add_argument('-d','--dryrun', dest='dryrun', action='store_true', help='Perform in dryrun mode')
    parser.add_argument('-t','--test', dest='test', action='store_true', help='Perform in test/staging mode')
    parser.add_argument('--fileset', dest='fileset',help='Fileset to add to fails index, concat multiple files with ","')
    parser.add_argument('-o','--output', dest='output', default=None, help='Send fail list to an output file')

    args = parser.parse_args()

    return {
        'conf': args.conf,
        'dryrun': args.dryrun,
        'test': args.test,
        'fileset': args.fileset,
        'output': args.output
    }

def dump_errors():
    """
    Dump all the failed files in the ES index into a local file
    """

    args = get_args()
    conf = load_config(args['conf'])
    outfile = args['output']

    esconf = {
        'headers': {
            'x-api-key': conf['elasticsearch']['x-api-key']
        },
            'retry_on_timeout': True,
            'timeout': 30
    }

    if args['test']:
        index = conf['failure_index_test']['name']
    else:
        index = conf['failure_index']['name']
    
    es = Elasticsearch(
        hosts=['https://elasticsearch.ceda.ac.uk'],
        headers=esconf['headers'])

    hits = es.search(index=index, doc_type="_doc")['hits']['hits']

    logger.info(f'Discovered {len(hits)} previous files')

    output = []
    for hit in hits:
        file = hit['_source']['info']['filename']
        output.append(file)
        id = hashlib.sha1(file.encode(errors="ignore")).hexdigest()

        es.delete(
            index=index,
            id=id
        )

    logger.info(f'Sending failed filelist to {outfile}')
    with open(outfile,'w') as f:
        f.write('\n'.join(output))

def add_errors():
    """
    Add all the failed files to an ES index for later retrieval.
    """
    
    args = get_args()

    conf = load_config(args['conf'])

    filenames = args['fileset'].split(',')
    if len(filenames) == 1:
        filenames = [filenames]
    
    esconf = {
        'headers': {
            'x-api-key': conf['elasticsearch']['x-api-key']
        },
            'retry_on_timeout': True,
            'timeout': 30
    }
    if args['test']:
        index = conf['failure_test_index']['name']
    else:
        index = conf['failure_index']['name']
    
    es = CEDAElasticsearchClient(headers=esconf['headers'])
    
    for f in filenames:
        status = f.split('.')[0]

        with open(f) as g:
            content = [r.strip() for r in g.readlines()]

        for file in content:

            info = {
                'filename':file,
                'status':status
            }

            logger.info(f'{file}: {status} - uploaded')

            id = hashlib.sha1(file.encode(errors="ignore")).hexdigest()

            es.update(
                index=index,
                id=id,
                body={'doc': {'info':info}, 'doc_as_upsert': True}
            )



# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '11 Apr 2025'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

from pathlib import Path
from pathlib import _ignore_error as pathlib_ignore_error
from cci_os_worker import logstream
import logging
import asyncio
import os
from typing import Union
import re
import argparse
import aiofiles.os as aos
import glob
import json

from .utils import set_verbose, load_config

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

SYMLINK = None
DEPOSIT = None

def check_valid_path(path):
    """
    Check that we have been given a real directory
    :param path:
    :return: boolean
    """
    if path == '/':
        raise Exception('Cannot scan from root')

    if not bool(os.path.exists(path) and os.path.isdir(path)):
        raise OSError('{} is not a directory'.format(path))
    
def walk_storage_links(path: str, depth: int = 0, max_depth: int = None):
    """
    Used within the archive to follow links to storage pots but ignore links which are
    back within the archive and could be circular.
    :param path:
    :param depth:
    :param max_depth:
    :return:
    """
    top = os.fspath(path)
    dirs = []
    nondirs = []

    # We may not have read permission for top, in which case we can't
    # get a list of the files the directory contains.  os.walk
    # always suppressed the exception then, rather than blow up for a
    # minor reason when (say) a thousand readable directories are still
    # left to visit.  That logic is copied here.
    try:
        scandir_it = os.scandir(top)
    except OSError:
        return

    if max_depth:
        if depth >= max_depth:
            return

    with scandir_it:
        while True:
            try:
                try:
                    entry = next(scandir_it)
                except StopIteration:
                    break
            except OSError as error:
                logger.error(error)
                return

            try:
                is_dir = entry.is_dir()
            except OSError:
                # If is_dir() raises an OSError, consider that the entry is not
                # a directory, same behaviour than os.path.isdir().
                is_dir = False

            if is_dir:
                dirs.append(entry.name)
            else:
                nondirs.append(entry.name)

    # Yield before recursion when going top down
    yield top, dirs, nondirs

    depth += 1
    # Recurse into sub-directories
    islink, join = os.path.islink, os.path.join
    for dirname in dirs:
        new_path = join(top, dirname)
        if islink(new_path):
            # Only follow links to storage locations
            if os.readlink(new_path).startswith('/datacentre'):
                yield from walk_storage_links(new_path, depth, max_depth)
        else:
            # If the path is not a link, recurse
            yield from walk_storage_links(new_path, depth, max_depth)

class RescanDirs:

    def __init__(
            self, 
            scan_path: str,
            scan_level: int = 1,
            use_rabbit: bool = False,
            conf: str = '',
            dryrun: bool = False, 
            skip_dirs: bool = False,
            skip_files: bool = False,
            recursive: bool = False,
            file_regex: Union[str,None] = None,
            extension: Union[str,None] = None,
            output: str = None
        ) -> None:

        if scan_path == '':
            self._init_from_args()
            return
        
        if scan_level == 1:
            # Directory level search
            check_valid_path(scan_path)

            self.scan_path = os.path.abspath(scan_path)
        else:
            self.scan_path = scan_path

        self.scan_level = scan_level
        self.use_rabbit = use_rabbit
        self.conf = load_config(conf)

        self.file_limit = self.conf.get('file_limit',None)

        # Default match any filename not starting with a . dot
        self._file_regex = file_regex
        self._extension  = extension

        self._dryrun = dryrun
        self._recursive = recursive
        self._output = output

        self.skip_dirs = skip_dirs
        self.skip_files = skip_files

        self.routing_key = 'elasticsearch_update_queue_opensearch_ingest'

    @property
    def file_regex(self):
        if self._file_regex is not None and self._extension is not None:
            regex = f'{self._file_regex}(.{self._extension})$'
            try:
                re.compile(regex)
                return regex
            except re.error:
                raise ValueError(
                    'Incompatible regex and extensions given - '
                    f'{regex} is not valid regular expression.'
                )
        if self._extension is not None:
            return f'.+?(.{self._extension})$'
        elif self._file_regex is not None:
            return self._file_regex
        else:
            return '.+'


    @property
    def max_depth(self):
        if self._recursive:
            return None
        else:
            return 1

    def _init_from_args(self):

        default_config = os.path.join(os.path.dirname(__file__), '../conf/rabbit_updater.ini')

        parser = argparse.ArgumentParser(description='Submit directories/items to be re-scanned.')
        parser.add_argument('dir', type=str, help='Directory to scan')
        parser.add_argument('-r', dest='recursive', action='store_true',
                            help='Recursive. Will include all directories below this point as well')

        parser.add_argument('-l','--scan-level',type=int, dest='scan_level',
                            help='Level of depth for scanning (1,2,3)')
        parser.add_argument('-R','--use-rabbit',dest='use_rabbit',
                            help='Deposit to rabbit queues or return list of paths')
        parser.add_argument('-v','--verbose', action='count', default=2, help='Set level of verbosity for logs')

        #parser.add_argument('--no-files', dest='nofiles', action='store_true', help='Ignore files')
        
        # Removed the ability to publish whole directories
        #parser.add_argument('--no-dirs', dest='nodirs', action='store_true', help='Ignore directories')
        parser.add_argument('--conf', type=str, default=default_config, help='Optional path to configuration file')
        parser.add_argument('--dry-run', dest='dryrun', action='store_true', help='Display log messages to screen rather than pushing to rabbit')

        parser.add_argument('-o','--output',dest='output', help='Store output list in a file.')

        parser.add_argument('--file-regex', dest='file_regex', 
                            help='Matching file regex, by default regex applies to all files not starting with "."',
                            default=None)
        parser.add_argument('--extension', dest='extension', 
                            help='Matching files by file extension.', default=None)
        args = parser.parse_args()

        set_verbose(args.verbose)

        self.__init__(
            args.dir,
            scan_level=args.scan_level,
            use_rabbit=args.use_rabbit,
            conf=args.conf,
            dryrun=args.dryrun,
            recursive=args.recursive,
            file_regex=args.file_regex,
            output=args.output,
            extension=args.extension
        )

    def _setup_rabbit(self):

        raise NotImplementedError(
            'Rabbit Queues have not been migrated to the CCI OS Worker package.' \
            'Please contact daniel.westwood@stfc.ac.uk for details on reimplementation if needed.'
        )

    def _submit_to_rabbit(self, item: str, itype = 'DEPOSIT') -> None:
        """
        Perform all operations for a specific file.
        All checks in relation to filepath should be checked
        before this stage.
        """
        raise NotImplementedError(
            'Rabbit Queues have not been migrated to the CCI OS Worker package.' \
            'Please contact daniel.westwood@stfc.ac.uk for details on reimplementation if needed.'
        )

        logger.info(f'Depositing {item} to Rabbit')

        msg = self.rabbit_connection.create_message(item, itype) #Deposit
        self.rabbit_connection.publish_message(msg, routing_key=self.routing_key) #'opensearch.tagger.cci')

    def _determine_paths(self):
        """
        Obtain the list of filepaths to enter
        into the facet scanner.

        This is either based on a file path, gathering
        all files under a directory (with a given regex),
        or based on a submission of JSON files.
        """

        scan_files = []

        if self.scan_level == 2: # All files under a directory
            logger.info('Scanning directories')
            for root, dirs, files in walk_storage_links(self.scan_path, max_depth=self.max_depth):
                for file in files:
                    if not re.match(self.file_regex, file):
                        continue

                    scan_files.append(f'{root}/{file}')

        else:
            # Pull files from json
            logger.info(f'Scanning JSON directory: {self.scan_path}')
            scanpath = f'{os.path.abspath(self.scan_path)}/**/*.json'
            jsons = glob.glob(scanpath, recursive=True)

            if self.file_limit:
                if len(jsons) != 0:
                    self.file_limit = int(self.file_limit/len(jsons))
                else:
                    raise FileNotFoundError(
                        'No JSON files were located.'
                    )
            else:
                self.file_limit = 99999999999999

            for js_count, file in enumerate(jsons):
                logger.info(f'Processing {file}')
                # Only want to track the changes in the JSON directory
                if not file.endswith('.json'):
                    continue

                with open(file) as reader:
                    data = json.load(reader)
                if 'datasets' not in data:
                    logger.warning(f'File {file}: missing "datasets" attribute')
                    continue
                ds = data['datasets']

                if not hasattr(ds, '__iter__'):
                    logger.warning(f'File {file}: "datasets" property is not iterable.')
                    continue

                add_files = []

                dfiles = []
                for ds_count, d in enumerate(ds):
                    # Find all single files
                    dfiles = [f for f in glob.glob(f'{d}/**/*.*', recursive=True) if re.match(self.file_regex,f)]
                    add_files += dfiles

                    logger.info(f' > (j: {js_count+1}/{len(jsons)}, d: {ds_count+1}/{len(ds)})')
                    logger.info(f' > {len(dfiles)} datasets ({file.split("/")[-1]}) ({len(scan_files)} total)')

                    if len(add_files) > self.file_limit:
                        add_files = add_files[:self.file_limit]
                        break

                scan_files += add_files

                logger.info(f'Added {len(add_files)} files from JSON file {js_count+1}')

        return scan_files

    def scan(self) -> list:
        output_files = 0

        if self.use_rabbit:
            self._setup_rabbit()

        deposit_paths = []

        for path in self._determine_paths():
            # Note the mkdir and symlink messages are no longer
            # required as all files have been ingested separately.

            output_files += 1

            # Create symlink message for file links
            if os.path.islink(path):
                action = SYMLINK
            else:
                action = DEPOSIT

            if self._dryrun:
                logger.info(f'{action}: {path}')
                continue

            if self.use_rabbit:
                self._submit_to_rabbit(path, itype=action)
            else:
                # Do something with the paths here.
                deposit_paths.append(path)

        logger.info(f'Submitted {output_files} files')
        return deposit_paths

    def save_data(self, outdata):

        if self._output is None:
            for line in outdata:
                print(line)
            return
        
        with open(self._output,'w') as f:
            f.write('\n'.join(outdata))

def check_timeout():

    async def path_exists(path) -> bool:
        try:
            await aos.stat(str(path))
        except OSError as e:
            if not pathlib_ignore_error(e):
                raise
            return ''
        except ValueError:
            # Non-encodable path
            return ''
        return True

    async def listfile():
        async with asyncio.timeout(10):
            await path_exists('/neodc/esacci/esacci_terms_and_conditions.txt')
    try:
        status = asyncio.run(listfile())
    except TimeoutError:
        logger.error('TIMEOUT: ESACCI Directories inaccessible')
        return True

    # If we didn't get a timeout error, can now perform a standard check.
    if not os.path.isfile('/neodc/esacci/esacci_terms_and_conditions.txt'):
        logger.error('INACCESSIBLE: ESACCI Directories inaccessible')
        return True
    return False

def rescan_directory():

    logger.info("Starting rescan check")
    if check_timeout():
        return
    logger.info("Archive access check: SUCCESS")

    r = RescanDirs('')
    if not r.use_rabbit:
        r.save_data(r.scan())
    else:
        _ = r.scan()


# Logger setup
import logging

logging.basicConfig(level=logging.DEBUG)
logstream = logging.StreamHandler()

formatter = logging.Formatter('%(levelname)s [%(name)s]: %(message)s')
logstream.setFormatter(formatter)



import os
import datetime

from .util import calculate_md5


class GenericFile(object):
    """
    Class for returning basic information about a file.
    """

    LEVEL_MAP = {
        "1": 'get_metadata_level1',
        "2": 'get_metadata_level2',
        "3": 'get_metadata_level3',
    }

    def __init__(self, file_path, level, calculate_md5=False):
        self.file_path = file_path
        self.level = str(level)
        self.handler_id = None
        self.calculate_md5 = calculate_md5

    def _get_file_ownership(self):

        uid = os.stat(self.file_path).st_uid
        gid = os.stat(self.file_path).st_gid

        return uid, gid

    def get_metadata_level1(self):
        """
        Scans the given file and returns information about 
        the file not the content.
        :returns: A dict containing summary information.
        """

        self.handler_id = "Generic level 1."

        #Do the basic checking, if file exists 
        #and that it is not a symbolic link.
        if ( self.file_path is None
             or not os.path.isfile(self.file_path)
             # or os.path.islink(self.file_path)
           ):
            return None, None, None

        file_info = {}
        info = {}

        file_stats = os.stat(self.file_path)

        #Basic information. 
        info["name"] = os.path.basename(self.file_path) #ntpath.basename(file_path)
        info["name_auto"] = info["name"]
        info["directory"] = os.path.dirname(self.file_path)
        info["location"] = "on_disk"

        uid,gid = self._get_file_ownership()
        info["user"] = uid
        info["group"] = gid

        info["is_link"] = os.path.islink(self.file_path)

        info["last_modified"] = datetime.datetime.fromtimestamp(file_stats.st_mtime).isoformat()

        info["size"] = os.path.getsize(self.file_path)

        file_type = os.path.splitext(info["name"])[1]
        if len(file_type) == 0:
            file_type = "File without extension."

        info["type"] = file_type

        if self.calculate_md5:
            info["md5"] = calculate_md5(self.file_path)

        if getattr(self, 'FILE_FORMAT', None):
            info["format"] = self.FILE_FORMAT

        file_info["info"] = info
        return file_info, None, None

    def get_metadata_level2(self):

        """
         Wrapper for method get_properties_generic_level1().
        :returns: A dict containing information compatible with current es ops.
        """

        file_info, a, b = self.get_metadata_level1()

        self.handler_id = "Generic level 2."

        if file_info is None:
            return None, None, None

        return file_info, None, None

    def get_metadata_level3(self):
        file_info, a, b = self.get_metadata_level2()
        self.handler_id = "Generic level 3."
        return file_info, None, None

    def get_metadata(self):

        metadata_function = self.LEVEL_MAP.get(self.level)

        if getattr(self, metadata_function, None):
            return getattr(self, metadata_function)()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


if __name__ == "__main__":
    import datetime
    import sys

    # run test
    try:
        level = str(sys.argv[1])
        file = sys.argv[2]
    except IndexError:
        level = '1'
        file = '/badc/mst/data/nerc-mstrf-radar-mst/v4-0/st-mode/cardinal/2015/09/nerc-mstrf-radar-mst_capel-dewi_20150901_st300_cardinal_33min-smoothing_v4-0.nc'

    gf = GenericFile(file, level)
    start = datetime.datetime.today()
    print(gf.get_metadata())
    end = datetime.datetime.today()
    print(end - start)



import netCDF4
import six
from dateutil.parser import parse
import re
import logging
import traceback

from .generic_file import GenericFile
import cci_os_worker.filehandlers.util as util
import cci_os_worker.filehandlers.geojson as geojson

logger = logging.getLogger(__name__)


def time_order(time1, time2):
    """
    Make sure that the times are the correct way round
    where start time is before end time.

    :param time1: First time input
    :param time2: Second time input

    :return: start_time, end_time
    """
    start_time = time1 if time1 < time2 else time2
    end_time = time2 if time2 > time1 else time1

    return start_time, end_time


def sanitise_float(value):
    """
    Clean text input for the lat and lon values
    Make sure that we return a float value.

    Some cases the extracted value is of the form: -90.0f; // float
    :param value:
    :return:
    """

    try:
        return float(value)
    except ValueError:
        # Clean all non-numeric characters
        value = re.sub('[^-.\d]', '', str(value))

    return float(value)


class NetCdfFile(GenericFile):
    """
    Simple class for returning basic information about the content
    of an NetCDF file.
    """

    def __init__(self, file_path, level, additional_param=None, **kwargs):
        GenericFile.__init__(self, file_path, level, **kwargs)
        self.FILE_FORMAT = "NetCDF"
        self.es = additional_param

    @staticmethod
    def clean_coordinate(coord):
        """Return True if coordinate is valid."""
        try:
            # This filters out misconfigured "_FillValue" elements
            if coord == 0.0:
                return False

            int(coord)  # If this fails, "coord" is not a number!

            return True
        except ValueError:
            return False

    def geospatial(self, ncdf, lat_name, lon_name):
        """
        Return a dict containing lat/lons from NetCDF file.

        :param Dataset ncdf: Reference to an opened netcdf4.Dataset object
        :param lat_name: Name of parameter containing latitude values
        :param lon_name: Name of parameter containing longitude values
        :returns: Geospatial information as dict.
        """

        lat_min = getattr(ncdf, 'geospatial_lat_min', None)
        lat_max = getattr(ncdf, 'geospatial_lat_max', None)
        lon_min = getattr(ncdf, 'geospatial_lon_min', None)
        lon_max = getattr(ncdf, 'geospatial_lon_max', None)

        if all([lat_min, lat_max, lon_min, lon_max]):
            return {
                "type": "track",
                "lat": [sanitise_float(lat_min), sanitise_float(lat_max)],
                "lon": [sanitise_float(lon_min), sanitise_float(lon_max)]
            }

        lats = ncdf.variables[lat_name][:].ravel()
        lons = ncdf.variables[lon_name][:].ravel()
        return {
            "type": "track",
            "lat": lats,
            "lon": lons
        }

    def find_var_by_standard_name(self, ncdf, standard_name):
        """
        Find a variable reference searching by CF standard name.

        :param Dataset ncdf: Reference to an opened netCDF4.Dataset object
        :param str standard_name: The CF standard name to search for
        """
        for key, value in six.iteritems(ncdf.variables):
            try:
                if value.standard_name.lower() == standard_name.lower():
                    return key
            except AttributeError:
                continue

    def get_geospatial(self, ncdf):
        lat_name = self.find_var_by_standard_name(ncdf, "latitude")
        lon_name = self.find_var_by_standard_name(ncdf, "longitude")

        if lat_name and lon_name:
            return self.geospatial(ncdf, lat_name, lon_name)

    def get_temporal(self, ncdf):

        temporal = {}

        time_name = self.find_var_by_standard_name(ncdf, "time")
        times = None

        # Try time coverage attributes
        start_time = getattr(ncdf, 'time_coverage_start', None)
        end_time = getattr(ncdf, 'time_coverage_end', None)

        # If extracting from the header not successful, try to retrieve the time
        # coordinate, if we have a coordinate name
        if not all([start_time, end_time]) and time_name:
            try:
                times = list(netCDF4.num2date(list(ncdf.variables[time_name]),
                                              ncdf.variables[time_name].units))
            except AttributeError:
                pass

        # retrieve the start and end times
        if start_time:
            start_time = parse(start_time)
        elif times:
            start_time = times[0]

        if end_time:
            end_time = parse(end_time)
        elif times:
            end_time = times[-1]

        # Check the order if we have a start and end
        if all([start_time, end_time]):
            start_time, end_time = time_order(start_time, end_time)

        if start_time and not end_time:
            end_time = start_time

        # Build the response
        if start_time:
            temporal = {
                "time_range": {
                    "gte": start_time.isoformat()
                },
                "start_time": start_time.isoformat()
            }

        # Need to use get because it is possible that a start date
        # was not found
        if end_time:
            time_range = temporal.get('time_range', {})
            time_range.update({
                "lte": end_time.isoformat()
            })
            temporal["end_time"] = end_time.isoformat()

        return temporal

    def get_phenomena(self, netcdf):
        """
        Construct list of Phenomena based on variables in NetCDF file.
        :returns : List of metadata.product.Parameter objects.
        """
        phen_list = []

        # for all phenomena list.
        for v_name, v_data in six.iteritems(netcdf.variables):
            new_phenomenon = {}
            phen_attr_list = []

            # for all attributtes in phenomenon.
            for key, value in six.iteritems(vars(v_data)):

                if not util.is_valid_phenomena(key, value):
                    continue

                phen_attr = dict(
                    name=str(key.strip()),
                    value=(value.strip()).encode('utf-8', 'ignore').decode()
                )
                phen_attr_list.append(phen_attr)

            # Add the variable ID
            phen_attr = dict(name='var_id', value=str(v_name))
            phen_attr_list.append(phen_attr)

            # Extract extra information about each variable not in the __dict__ response
            for attr in ["dtype", "shape", "_ChunkSizes", "_FillValue", "dimensions", "size"]:
                try:
                    value = getattr(v_data, attr)

                    if attr == 'dtype':
                        value = value.name

                    # Convert value datatype to JSON compatible
                    if type(value) not in (str, tuple, list):
                        value = str(value)

                    phen_attr = dict(name=attr, value=value)
                    phen_attr_list.append(phen_attr)

                except AttributeError:
                    pass

            if len(phen_attr_list) > 0:
                new_phenomenon["attributes"] = phen_attr_list

                phen_list.append(new_phenomenon)

        file_phenomena = util.build_phenomena(phen_list)

        return file_phenomena

    def get_metadata_level2(self):

        file_info = self.get_metadata_level1()

        if file_info is not None:
            try:
                with netCDF4.Dataset(self.file_path) as netcdf_object:
                    netcdf_phenomena = self.get_phenomena(netcdf_object)
                    file_info[0]["info"]["read_status"] = "Successful"
                return file_info + netcdf_phenomena
            except Exception:
                file_info[0]["info"]["read_status"] = "Read Error"

                return file_info + (None,)

    def get_metadata_level3(self):

        """
        Wrapper for method phenomena().
        :returns:  A dict containing information compatible with current es index level 2.
        """
        # level 1
        file_info = self.get_metadata_level1()[0]
        spatial = None
        netcdf_phenomena = None

        if file_info is None:
            return None

        with netCDF4.Dataset(self.file_path) as netcdf:
            netcdf_phenomena = self.get_phenomena(netcdf)[0]
            geo_info = self.get_geospatial(netcdf)
            temp_info = self.get_temporal(netcdf)

        self.handler_id = "Netcdf handler level 3."

        if geo_info:

            spatial = None

            gj = geojson.GeoJSONGenerator(geo_info["lat"], geo_info["lon"])
            spatial = gj.get_elasticsearch_geojson()
        if spatial:
            spatial = {"coordinates": spatial["geometries"]["search"]}

        if temp_info:
            file_info["info"]["temporal"] = temp_info

            file_info["info"]["read_status"] = "Successful"
        return file_info, netcdf_phenomena, spatial

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


if __name__ == "__main__":
    import datetime
    import sys

    # run test
    try:
        level = str(sys.argv[1])
        file = sys.argv[2]
    except IndexError:
        level = '1'
        file = '/badc/mst/data/nerc-mstrf-radar-mst/v4-0/st-mode/cardinal/2015/09/nerc-mstrf-radar-mst_capel-dewi_20150901_st300_cardinal_33min-smoothing_v4-0.nc'

    ncf = NetCdfFile(file, level)
    start = datetime.datetime.today()
    end = datetime.datetime.today()



"""
Module for holding and exporting file metadata as JSON documents.
"""

from __future__ import division
import math
import numpy.ma as ma
import numpy as np
from six.moves import xrange


class GeoJSONGenerator(object):
    """
    A class to translate between different scientific data types and GeoJSON.

        * Flight tracks => GeoJSON "LineString"
        * Photography / stationary observation points => GeoJSON "Point"
        * Satellite swaths => GeoJSON "MultiPolygon"
    """
    def __init__(self, latitudes, longitudes, shape_type=None):
        self._sanitise_geometry(ma.array(longitudes), ma.array(latitudes))
        self.shape_type = shape_type

    def get_elasticsearch_geojson(self):
        """
        Returns the specified GeoJSON object, constructed from the geometry
        supplied at construction-time. This can be provided from the
        "shape_type" parameter. Can be one of the following values:

            * None (returns a "LineString" type) [defaults to "track"]
            * "track" (returns a "LineString" type) [default]
            * "point" (returns a "Point" type)
            * "swath" (returns a "MultiPolygon" type)

        :return: Specified GeoJSON object.
        """
        geojson = None
        if len(self.latitudes) > 0 and len(self.longitudes) > 0:
            if (self.shape_type == "point" or
                    self._num_points(self.longitudes, self.latitudes) == 1):
                geojson = {
                    "geometries": {
                        "search": self._gen_point(),
                        "display": self._gen_point()
                    }
                }
            elif self.shape_type == "swath":
                geojson = {
                    "geometries": {
                       "search": self._gen_swath(),
                       "display": self._gen_track()
                    }
                }
            elif self.shape_type == "track":
                geojson = {
                    "geometries": {
                        "search": self._gen_envelope(),
                        "display": self._gen_track()
                    }
                }
            elif self.shape_type is None:
                geojson = {
                    "geometries": {
                        "search": self._gen_envelope()
                    }
                }

        return geojson

    def _num_points(self, lons, lats):
        """
        Return the number of coordinates in the SHORTEST list out of:
            * self.latitudes
            * self.longitudes

        This returns the length of the SHORTEST list.

        :return int: The length of the shortest coordinate list.
        """
        len_lons = np.size(lons)
        len_lats = np.size(lats)

        if len_lons > len_lats:
            return len_lats
        else:
            return len_lons

    def __align_lons_lats(self, lons, lats):
        """
        Ensure "lons" and "lats" are the same length.
        """
        size = self._num_points(lons, lats)
        if np.size(lons) > size:
            lons = lons[0:size]
        elif np.size(lats) > size:
            lats = lats[0:size]

        return (lons, lats)

    def _sanitise_geometry(self, lons, lats):
        """
        Sanitise geometry by removing any masked points.
        :param lons:
        :param lats:
        :return: A tuple containing arrays of (lon, lat)
        """
        # Align the arrays
        lons, lats = self.__align_lons_lats(lons, lats)

        # Get masks
        lon_mask = ma.getmaskarray(lons)
        lat_mask = ma.getmaskarray(lats)

        # Filter the arrays
        # kltsa 19/07/2016 Change for issue 23330: Some filters 
        #                  were excluded in order to allow values 
        #                  up to 90 to be included.
        self.longitudes = lons[
            (lons >= -180) &
            (lons <= 180) &
            #(lats >= 90) &
            #(lats <= 90) &
            (lon_mask == False)# &
            #(lat_mask == False)
        ]

        self.latitudes = lats[
            #(lons >= -180) &
            #(lons <= 180) &
            (lats >= -90) &
            (lats <= 90) &
            #(lon_mask == False) &
            (lat_mask == False)
        ]


    def _gen_point(self):
        """
        Returns a GeoJSON 'Point' type.

        :return: A GeoJSON 'Point' type in a dictionary.
        """
        geojson = {
            "type": "Point",
            "coordinates": [float(self.longitudes[0]), float(self.latitudes[0])]
         }

        return geojson

    def _gen_swath(self, num_polygons=30):
        """
        Returns a GeoJSON 'MultiPolygon' type.

        :return: A GeoJSON 'MultiPolygon' type in a dictionary.
        """
        geojson = {
            "type": "MultiPolygon",
            "coordinates": []
        }

        track = self._gen_track(num_polygons + 1)["coordinates"]
        for i in xrange(len(track) - 1):
            lo = track[i]
            hi = track[i + 1]

            polygon = self._gen_polygon(lo[0], hi[0], lo[1], hi[1])
            geojson["coordinates"].append(polygon["coordinates"])

        return geojson

    def _gen_track(self, num_points=30):
        """
        Creates a LineString from a summary of the file's latitudes and
        longitudes. Samples a coordinate every 'num_points'.

        :return: A GeoJSON LineString containing a sample of points from the
                  flight track.
        """
        track = {
            "type": "LineString"
        }

        point_count = self._num_points(self.longitudes, self.latitudes)
        if point_count <= num_points:
            track_lons = map(float, self.longitudes)
            track_lats = map(float, self.latitudes)
        else:
            step = int(math.ceil(point_count / num_points))
            track_lons = map(float,self.longitudes[::step])
            track_lats = map(float,self.latitudes[::step])

        track["coordinates"] = zip(track_lons, track_lats)

        return track

    def _gen_envelope(self):
        """
        Generate and return an Elasticsearch envelope type.
        :return: A bounding box as an envelope.
        """
        lon_left, lon_right = self._get_bounds(self.longitudes,
                                               wrapped_coords=True)

        lat_bottom, lat_top = self._get_bounds(self.latitudes)

        if (lon_left is None or lon_right is None or
                lat_bottom is None or lat_top is None):
            return None

        envelope = {
            "type": "envelope",
            "coordinates": [
                [round(lon_left, 3), round(lat_top, 3)],
                [round(lon_right, 3), round(lat_bottom, 3)]
            ]
        }

        return envelope

    def _gen_bbox(self):
        """
        Generate and return a bounding box for the given geospatial data.
        If there are no data points then bounding box is none

        :return: A bounding-box formatted as GeoJSON
        """
        lon_left, lon_right = self._get_bounds(self.longitudes,
                                               wrapped_coords=True)
        lat_bottom, lat_top = self._get_bounds(self.latitudes)

        if (lon_left is None or lon_right is None or
                lat_bottom is None or lat_top is None):
            return None

        bbox = self._gen_polygon(lon_left, lon_right, lat_bottom, lat_top)

        return bbox

    @staticmethod
    def _gen_polygon(lon_l, lon_r, lat_b, lat_t):
        """
        Generate and return a polygon from two provided sets of coordinates.

        :return: A GeoJSON bounding box
        """
        corners = [
            [
                map(float,[lon_r, lat_t]),
                map(float,[lon_l, lat_t]),
                map(float,[lon_l, lat_b]),
                map(float,[lon_r, lat_b]),
                map(float,[lon_r, lat_t])
            ]
        ]

        bbox = {
            "type": "polygon",
            "orientation": "counterclockwise",
            "coordinates": corners
        }

        return bbox

    @staticmethod
    def _get_bounds(item_list, wrapped_coords=False):
        """
        Return a tuple containing the first and secound bound in the list.
            * For 0 values it returns (None, None).
            * For 1 value it returns that value twice
            * For 2 values it returns those (low, high) for unwrapped
              co-ordinates or in the order given for wrapped
            * For Multiple values unwrapped, this returns min and max
            * For Multiple values wrapped co-ordinates, this
              returns the value around the largest gap in values

        :param list item_list: List of comparable data items
        :param wrapped_coords: is this a coordinate which wraps at 360 back to
                               0, e.g. longitude
        :return: Tuple of (first bound, second bound for values in the list)
        """
        items = ma.compressed(item_list)
        if len(items) < 1:
            return None, None

        if len(items) == 1:
            return items[0], items[0]

        if wrapped_coords:
            if len(items) == 2:
                first_bound_index = 0
                second_bound_index = 1
            else:
                # find the largest angle between closest points and exclude
                # this from the bounding box ensuring that this includes going
                # across the zero line
                items = sorted(items)
                first_bound_index = 0
                second_bound_index = len(items) - 1
                max_diff = (items[first_bound_index] -
                            items[second_bound_index]) % 360
                for i in range(1, len(items)):
                    diff = (items[i] - items[i-1]) % 360
                    if diff > max_diff:
                        max_diff = diff
                        first_bound_index = i
                        second_bound_index = i-1

            first_bound = items[first_bound_index]
            second_bound = items[second_bound_index]
        else:
            first_bound = min(items)
            second_bound = max(items)

        return float(first_bound), float(second_bound)





class TestConsistency:

    def test_core(self):

        from cci_os_worker.errors import HandlerError, DocMetadataError
        from cci_os_worker.all_facets import (
            FacetUpdateHandler,
            _get_command_line_args,
            main
        )

        from cci_os_worker.path_tools import PathTools

        from cci_os_worker.utils import (
            set_verbose,
            load_config,
            load_datasets,
            UpdateHandler
        )

        assert 1==1

    def test_netcdf(self):

        from cci_os_worker.filehandlers.generic_file import GenericFile
        from cci_os_worker.filehandlers.geojson import GeoJSONGenerator
        from cci_os_worker.filehandlers.netcdf_file import NetCdfFile
        from cci_os_worker.filehandlers.util import Parameter


# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '05 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

import logging
from cci_os_worker import logstream

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

class HandlerError(Exception):
    """A handler could not be identified for this file"""
    def __init__(self, filename = '<Not Given>', verbose = 0):
        self.message = f"A handler could not be identified for this file: {filename}"
        if verbose < 1:
            self.__class__.__module__ = 'builtins'
    def get_str(self):
        return 'HandlerError'

class DocMetadataError(Exception):
    """Unable to retrieve metadata from the handler"""
    def __init__(self, filename = '<Not Given>', verbose = 0):
        self.message = f"Unable to retrieve metadata from the handler for this file: {filename}"
        if verbose < 1:
            self.__class__.__module__ = 'builtins'
    def get_str(self):
        return 'DocMetadataError'


# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '05 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

import os
import yaml
import json
import logging

from cci_os_worker import logstream

from elasticsearch import Elasticsearch

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

def set_verbose(level: int):
    """
    Reset the logger basic config.
    """

    levels = [
        logging.WARN,
        logging.INFO,
        logging.DEBUG,
    ]

    if level >= len(levels):
        level = len(levels) - 1

    for name in logging.root.manager.loggerDict:
        lg = logging.getLogger(name)
        lg.setLevel(levels[level])

def load_config(conf: str):
    """
    Load a conf.yaml file to a dictionary
    """
    if os.path.isfile(conf):
        with open(conf) as f:
            config = yaml.safe_load(f)
        return config
    else:
        logger.error(f'Config file {conf} unreachable')
        return None

def load_datasets(datafile_path: str):
    if not os.path.isfile(datafile_path):
        raise ValueError(
            f'Filepath {datafile_path} is not accessible'
        )

    with open(datafile_path) as f:
        datasets = [r.strip() for r in f.readlines()]
    return datasets

class UpdateHandler:

    def __init__(self, conf: dict, dryrun: bool = False, test: bool = False):
        """
        Initialise this class with config and other switches.
        """

        self._dryrun = dryrun
        self._test = test
        self._conf = conf

        api_key = conf['elasticsearch']['x-api-key']

        self.es = Elasticsearch(
            hosts=['https://elasticsearch.ceda.ac.uk'],
            headers={'x-api-key': api_key})

    def _local_cache(self, filename, contents):
        """
        Cache contents of Opensearch record locally
        for test purposes."""

        if os.path.isfile(filename):
            try:
                with open(filename) as f:
                    refs = json.load(f)
                contents = contents | refs
            except:
                pass

        with open(filename,'w') as f:
            f.write(json.dumps(contents))

    def _warn_tagger_root(self):
        """
        Display warning message for unset tagger root
        """

        if not os.environ.get('JSON_TAGGER_ROOT'):
            logger.warning('!!!')
            logger.warning('JSON_TAGGER_ROOT environment variable is not set')
            logger.warning('!!!')

    def _ensure_cache(self):
        """
        Ensure the cache directory exists
        """

        if not os.path.isdir('cache/'):
            os.makedirs('cache/')

    def process_deposits(self, datafile_path: str, prefix: str = '', file_limit = None):
        """
        Process deposits from the dataset file given to this function
        """

        logger.info(f'Processing deposits for {datafile_path}')
        self._warn_tagger_root()

        if not os.path.isfile(datafile_path):
            raise ValueError(
                f'Filepath {datafile_path} is not accessible'
            )
        
        if self._dryrun:
            self._ensure_cache()

        with open(datafile_path) as f:
            datasets = [prefix + r.strip() for r in f.readlines()]

        if file_limit != None:
            datasets = datasets[:file_limit]

        total = len(datasets)

        fail_list = []
        # Process all files individually.
        for idx, fp in enumerate(datasets):
            status = self._process_file(fp, index=idx+1, total=total)
            if status != 0:
                logger.error(status)
                fail_list.append(fp)

        self._warn_tagger_root()
        return fail_list

    def process_removals(self, datafile_path: str):
        """
        Process removals from the dataset file given to this function
        """

        if not os.path.isfile(datafile_path):
            raise ValueError(
                f'Filepath {datafile_path} is not accessible'
            )

        with open(datafile_path) as f:
            datasets = [r.strip() for r in f.readlines()]

        fail_list = []
        # Process all files individually.
        for fp in datasets:
            status = self._remove_file(fp)
            if status != 0:
                logger.error(status)
                fail_list.append((fp, status))
        return fail_list

    def _process_file(self, filepath: str, index: int = None, total: int = None):
        """
        Safe processing of filepath
        """

        try:
            self._single_process_file(filepath, index=index, total=total)
            return 0
        except Exception as err:
            return err


# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '07 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

from pathlib import Path
import os
import requests
from json.decoder import JSONDecodeError
import json
import hashlib
from requests.exceptions import Timeout
from ceda_directory_tree import DatasetNode

from typing import Optional, Tuple, List

import logging
from cci_os_worker import logstream

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

class SpotMapping():
    """
    Downloads the spot mapping from the cedaarchiveapp.
    Makes two queryable dicts:
        - spot2pathmapping = provide spot and return file path
        - path2spotmapping = provide a file path and the spot will be returned
    """
    url = "http://cedaarchiveapp.ceda.ac.uk/cedaarchiveapp/fileset/download_conf/"
    spot2pathmapping = {}
    path2spotmapping = {}

    # Remove logging message when running script
    logging.getLogger("requests").setLevel(logging.WARNING)

    def __init__(self, test=False, spot_file=None, sep='='):

        logging.info("Initialising spots mapping with test: {} and spot file: {}".format(test, spot_file))

        if test:
            self.spot2pathmapping['spot-1400-accacia'] = "/badc/accacia"
            self.spot2pathmapping['abacus'] = "/badc/abacus"

        elif spot_file:
            with open(spot_file) as reader:
                spot_mapping = reader.readlines()

            self._build_mapping(spot_mapping, sep=sep)

        else:
            self._download_mapping()

    def __iter__(self):
        return iter(self.spot2pathmapping)

    def __len__(self):
        return len(self.spot2pathmapping)

    def _download_mapping(self):
        """
        Download the mapping from the cedaarchiveapp and build mappings.
        """

        logging.info("Downloading spots from {}".format(self.url))

        response = requests.get(self.url)

        if response.status_code != 200:
            logging.error("Error getting mapping from cedaarchiveapp status code: {}, reason: {}".format(response.status_code, response.reason))
            raise Exception("bad response status code: {}, reason: {}".format(response.status_code, response.reason))

        spot_mapping = response.text.split('\n')

        self._build_mapping(spot_mapping)

    def _build_mapping(self, spot_mapping, sep=None):
        """
        Build the spot mapping dictionaries
        :param spot_mapping: list of mappings
        """

        for line in spot_mapping:
            if not line.strip(): continue
            spot, path = line.strip().split(sep)

            if spot in ("spot-2502-backup-test",): continue
            self.spot2pathmapping[spot] = path
            self.path2spotmapping[path] = spot

    def get_archive_root(self, key):
        """

        :param key: Provide the spot
        :return: Returns the directory mapped to that spot
        """

        # Get the path to the spot
        archive_root = self.spot2pathmapping.get(key)

        return archive_root

    def get_spot(self, key):
        """
        The directory stored in elasticsearch is the basename for the specific file. The directory stored on the spots
        page is further up the directory structure but there is no common cut off point as it depends on how many files there
        are in each dataset. This function recursively starts at the end of the directory stored in elasticsearch
        and gradually moves back up the file structure until it finds a match in the path2spot dict.

        :param key: Provide a filename or directory
        :return: Returns the spot which encompasses that file or directory.
        """

        archive_path = self.get_archive_path(key)

        if archive_path:
            while (archive_path not in self.path2spotmapping) and (archive_path != '/'):
                archive_path = os.path.dirname(archive_path)

        if archive_path == '/' or archive_path is None:
            return None

        return self.path2spotmapping[archive_path]

    def get_spot_from_storage_path(self, path):
        """
        Extract the spot name from the storage path

        :param path: Path to test
        :return: spot name and path suffix
        """

        # Setup output
        spot, suffix = None, None

        try:
            storage_suffix = path.split('/archive/')[1]
            spot = storage_suffix.split('/')[0]
            suffix = storage_suffix.split(spot + '/')[1]

        except IndexError:
            logging.warning("Error getting spot from: {}".format(path))

        return spot, suffix

    def get_archive_path(self, path):
        storage_path = os.path.realpath(path)

        spot, suffix = self.get_spot_from_storage_path(storage_path)

        spot_path = self.get_archive_root(spot)

        try:
            archive_path = os.path.join(spot_path, suffix)

        # Joining None produces an AttributeError in py2 and a TypeError py3
        except (AttributeError, TypeError):
            archive_path = spot_path

        return archive_path

    def is_archive_path(self, path):
        """
        Archive path refers to the location in the archive where the file exists.
        In the archive there are 3 path types:
            - storage path: the actual location of the file eg. /datacentre/archvol3/pan125/archive/namblex/data/...
            - archive path: the path with the spot mapping replacing the storage prefix eg. /badc/namblex/data/aber-radar-1290mhz/20020831/
            - symlink path: an alternative route to the file as displayed using pydap eg. badc/ncas-observations/data/man-radar-1290mhz/previous-versions/2002/08/31/

        This function takes the input path, gets the storage path, replaces the storage prefix with spot mapping to get the archive
        path and compares it to the input path. Returns True if input path is archive path.

        :param path: file path to test
        :return: Bool
        """

        return path == self.get_archive_path(path)

def process_observations(results):
    """
    Convert the result list into a mapping object
    :param results: list of observation json objects
    :return: object map
    """

    processed_map = {}
    for result in results:

        # Skip results where the publication state is working
        if result.get("publicationState") == "working":
            continue

        # Skip where the result_field value is None
        if result["result_field"] is None:
            continue

        data_path = result["result_field"]["dataPath"].rstrip("/")

        try:
            processed_map[data_path] = {
                "title": result["title"],
                "url": f'https://catalogue.ceda.ac.uk/uuid/{result["uuid"]}',
                "record_type": "Dataset",
            }
        except TypeError:
            continue

    return processed_map

def generate_moles_mapping(api_url, mapping=None):
    """
    Use the MOLES v2 API to generate a mapping from dataset path to moles record

    :param api_url: MOLES api URL
    :param mapping: Used for recursive functionality
    :return: Mapping dict
    """

    # Set the dictionary on first calling
    if not mapping:
        mapping = {}

    # Get the api response
    try:
        response = requests.get(api_url).json()
    except JSONDecodeError as e:
        import sys

        raise ConnectionError(
            f"Could not connect to {api_url} to get moles mapping"
        ) from e

    # Turn response into mapping object
    mapping.update(process_observations(response["results"]))

    if not response["next"]:
        return mapping
    else:
        return generate_moles_mapping(response["next"], mapping)

def load_moles_mapping(mapping_file):
    """
    Load a json document and condition it to match same as from API
    """

    data = {}

    with open(mapping_file) as reader:
        json_doc = json.load(reader)

    # Loop through and remove trailing slash from paths
    for k, v in json_doc.items():
        data[k.rstrip("/")] = v

    return data

class PathTools:
    def __init__(
        self,
        moles_mapping_url: str = "http://api.catalogue.ceda.ac.uk/api/v2/observations.json/",
        mapping_file: Optional[str] = None,
        spot_file: str = None,
    ):

        if os.path.isfile('moles_mapping.json'):
            mapping_file = 'moles_mapping.json'

        self.spots = SpotMapping(spot_file=spot_file)

        self.moles_mapping_url = moles_mapping_url

        if mapping_file:
            self.moles_mapping = load_moles_mapping(mapping_file)
        else:
            self.moles_mapping = generate_moles_mapping(self.moles_mapping_url)

        #DEBUG
        #with open('moles_mapping.json','w') as f:
            #f.write(json.dumps(self.moles_mapping))

        # Setup the matching tree
        self.tree = DatasetNode()
        for path in self.moles_mapping:
            self.tree.add_child(path)

    def generate_path_metadata(
        self, path: str
    ) -> Tuple[Optional[dict], Optional[bool]]:
        """
        Take path and process it to generate metadata as used in ceda directories index
        :param path: path to retrieve metadata for
        :return:
        """

        path = Path(path)

        try:
            if not path.exists():
                return None, None
        except PermissionError:
            return None, None

        # See if the path is a symlink and directory
        link = path.is_symlink()
        isdir = path.is_dir()

        # Set the archive path
        archive_path = path

        # If the path is a link, we need to find the path to the actual data
        if link:
            link_path = os.readlink(path)
            if not link_path.startswith(("/datacentre", "..")):
                archive_path = link_path
            elif link_path.startswith(".."):
                count = link_path.count("../")
                link_path = link_path.lstrip("./")
                archive_path = path.parents[count] / link_path

        # Create the metadata
        meta = {
            "depth": len(path.parts) - 1,
            "dir": path.name,
            "path": str(path),
            "archive_path": str(archive_path),
            "link": link,
            "type": "dir" if isdir else "file",
        }

        # Retrieve the appropriate MOLES record
        if isdir:
            record = self.get_moles_record_metadata(str(path))

            # If a MOLES record is found, add the metadata
            if record and record["title"]:
                meta["title"] = record["title"]
                meta["url"] = record["url"]
                meta["record_type"] = record["record_type"]

        return meta, meta["link"]

    def get_moles_record_metadata(self, path: str) -> Optional[dict]:
        """
        Try and find metadata for a MOLES record associated with the path.

        :param path: Directory path
        :return: Dictionary containing MOLES title, url and record_type
        """

        # Condition path - remove trailing slash
        path = path.rstrip("/")

        # Search the tree
        match = self.tree.search_name(path)
        if match:
            result = self.moles_mapping.get(path)

            if result:
                return result

        return self._get_moles_record_metadata_data_from_api(path)

    def _get_moles_record_metadata_data_from_api(self, path: str) -> Optional[dict]:
        """
        Request metadata from the API, this is used as a last resort

        :param path: Path to retrieve metadata for
        :return: Metadata dict | None
        """

        url = f"http://api.catalogue.ceda.ac.uk/api/v0/obs/get_info{path}"
        try:
            response = requests.get(url, timeout=10)
        except Timeout:
            return

        # Update moles mapping
        if response:
            self.moles_mapping[path] = response.json()
            return response.json()

    @staticmethod
    def get_readme(path: str) -> Optional[str]:
        """
        Search in directory for a README file and read the contents
        :param path: Directory path
        :return: Readme contents
        """
        if not os.path.exists(path):
            return

        if "00README" in os.listdir(path):
            with open(os.path.join(path, "00README"), errors="replace") as reader:
                content = reader.read()

            return content.encode(errors="ignore").decode()

    def update_mapping(self) -> bool:

        successful = True
        # Update the moles mapping
        try:
            self.spots._download_mapping()
            self.moles_mapping = requests.get(self.moles_mapping_url, timeout=30).json()
        except (ValueError, Timeout):
            successful = False

        return successful

    @classmethod
    def generate_id(cls, path: str) -> str:
        """
        Take a path, encode to utf-8 (ignoring non-utf8 chars) and return hash

        :param path: filepath to hash
        :return: hash hexdigest of sha1 hash
        """

        return hashlib.sha1(path.encode(errors="ignore")).hexdigest()


# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '05 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

import logging
import hashlib
import argparse
import os

from typing import Union

from cci_facet_scanner.core.facet_scanner import FacetScanner
from ceda_elasticsearch_tools.elasticsearch import CEDAElasticsearchClient
from fbi_directory_check.utils import check_timeout

from .utils import load_config, UpdateHandler, set_verbose

from cci_os_worker import logstream

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

class ElasticsearchDeleter(UpdateHandler):

    def __init__(self, conf: dict, dryrun: bool = False, test: bool = False):
        """
        Initialise this class with the correct connections to 
        establish an elasticsearch client.
        """
        logger.info('Loading ES Deleter')

        super().__init__(conf, dryrun=dryrun, test=test)

    def _remove_file(self, filepath):
        """
        Delete all records for a specific dataset
        """

        index = self._conf['facet_files_index']['name']

        id = hashlib.sha1(filepath.encode(errors="ignore")).hexdigest()

        if not self._dryrun:
            try:
                self.es.delete(
                    index=index,
                    id=id,
                )
                return 0
            except:
                return 1
        else:
            logger.info(f'DRYRUN: Skipped deleting {filepath.split("/")[-1]}')
            return 0
        
def _get_command_line_args():
    """
    Get the command line arguments for the facet scan
    """
    parser = argparse.ArgumentParser(description='Entrypoint for the CCI OS Worker on the CMD Line')
    parser.add_argument('datafile_path', type=str, help='Path to the "datasets.txt" file')
    parser.add_argument('conf', type=str, help='Path to Yaml config file for Elasticsearch')

    parser.add_argument('-d','--dryrun', dest='dryrun', action='store_true', help='Perform in dryrun mode')
    parser.add_argument('-t','--test', dest='test', action='store_true', help='Perform in test/staging mode')
    parser.add_argument('-v','--verbose', action='count', default=2, help='Set level of verbosity for logs')
    parser.add_argument('-o','--output', dest='output', default=None, help='Send fail list to an output file')

    args = parser.parse_args()

    return {
        'datafile_path': args.datafile_path,
        'conf': args.conf,
        'dryrun': args.dryrun,
        'test': args.test,
        'verbose': args.verbose-1,
        'output': args.output
    }

def main(args: dict = None):
    if args is None:
        args = _get_command_line_args()
    if isinstance(args['conf'], str):
        conf = load_config(args['conf'])

    if conf is None:
        logger.error('Config file could not be loaded')
        return
    if not os.path.isfile(args['datafile_path']):
        logger.error(f'Inaccessible Datafile - {args["datafile_path"]}')
        return
    
    if check_timeout():
        logger.error('Check-timeout failed')
        return

    set_verbose(args['verbose'])

    deleter = ElasticsearchDeleter(conf, dryrun=args['dryrun'], test=args['test'])
    fail_list = deleter.process_removals(args['datafile_path'])

    logger.info('Failed items:')
    for f in fail_list:
        logger.info(f)

    if args['output'] is not None and fail_list != []:
        with open(args['output'],'w') as f:
            f.write('\n'.join(fail_list))

if __name__ == '__main__':
    main()


"""
Module containing useful functions for the command-line interfaces.
"""

import os
import errno
import sys
import subprocess
import json
import six
import re
import io
import datetime
from dateutil import parser
import hashlib
import logging
import ldap3
from ldap3.core.exceptions import LDAPSessionTerminatedByServerError
from typing import Optional, Union, List
from pwd import getpwuid
from grp import getgrgid

# Python 2/3 compatibility
if sys.version_info.major > 2:
    from configparser import RawConfigParser as ConfigParser
else:
    from ConfigParser import ConfigParser

log_levels = {"debug": logging.DEBUG,
              "info": logging.INFO,
              "warning": logging.WARNING,
              "error": logging.ERROR,
              "critical": logging.CRITICAL
              }

MAX_ATTR_LENGTH = 256

logger = logging.getLogger(__name__)


class Parameter(object):
    """
    Placeholder/wrapper class for metadata parameters

    :param str name: Name of variable/parameter
    :param dict other_params: Optional - Dict containing other param metadata
    """

    def __init__(self, name, other_params=None):
        self.items = []
        self.name = name

        # Other arbitrary arguments
        if other_params:
            for key, value in six.iteritems(other_params):
                self.items.append(
                    self.make_param_item(
                        key.strip(),
                        value.strip()
                    )
                )

    @staticmethod
    def make_param_item(name, value):
        """
        Convert a name/value pair to dictionary (for better indexing in ES)

        :param str name: Name of the parameter item (e.g. "long_name_fr", etc)
        :param str value: Value of the parameter item (e.g. "Radiance")
        :return: Dict containing name:value information
        """
        return {"name": name,
                "value": value}

    def get(self):
        """Return the list of parameter items"""
        return self.items

    def get_name(self):
        """Return the name of the phenomenon."""
        return self.name


class FileFormatError(Exception):
    """
    Exception to raise if there is a error in the file format
    """
    pass


class LotusRunner:
    """
    Class to handle running of tasks using the LOTUS scheduler
    """
    def __init__(self, queue='par-single'):
        self.queue = queue
        self.task_list = []

    def _run_tasks_in_lotus(self) -> None:
        """
        Submit all tasks in task list
        """

        for task in self.task_list:
            self._submit_job(task)

    def _submit_job(self, task: str) -> None:
        """
        Submit the job to LOTUS
        :param task: Task to submit
        """

        if self.queue == 'short-serial':
            wall_time = '24:00:00'
        else:
            wall_time = '48:00:00'

        command = f'sbatch -p {self.queue} -t {wall_time} -e lotus_errors/%j.err {task}'

        print(f'Executing command: {command}')

        subprocess.call(command, shell=True)

    def read_task_file(self, filename: str) -> None:
        """
        Read task file and add to task list
        :param filename: Path to file containing list of tasks
        """
        with open(filename) as reader:
            self.task_list = reader.read().splitlines()

    @staticmethod
    def remove_task_file(filename: str) -> None:
        """
        Delete the task file once the jobs have been submitted

        :param filename: File to remove
        """
        os.remove(filename)

    def run_tasks_in_lotus(self, task_list: list) -> None:
        """
        Run the tasks using the lotus scheduler

        :param task_list: List of tasks to run
        """

        self.task_list = task_list

        self._run_tasks_in_lotus()

    def run_tasks_file_in_lotus(self, filename: str) -> None:
        """
        Load the tasks from file and run in lotus
        scheduler

        :param filename: Path to file containing list of tasks
        """

        self.read_task_file(filename)

        self._run_tasks_in_lotus()

        self.remove_task_file(filename)


class LDAPIdentifier:
    """
    Provides interface to interact with LDAP and get user names
    and group names. The results are cached, as this information
    doesn't change, to reduce load on LDAP.
    """
    def __init__(self, **kwargs):
        """
        :param kwargs: ldap3 Connection kwargs
        """
        self.conn = ldap3.Connection(**kwargs)
        self.users = {}
        self.groups = {}

    def _process_result(self, key: str) -> Optional[str]:
        """
        Process LDAP response object and return the first value for the
        given key.

        :param result: LDAP response object
        :param key: key to return
        :return: value
        """
        if self.conn.entries:
            entry = self.conn.entries[0]
            return getattr(entry, key).value

    def _ldap_query(self, *args, **kwargs) -> None:
        """
        Wraps the LDAP search operation to catch errors
        caused by a closed connection. This method
        restarts the connection and tries the search again.

        :param args: args to pass to ldap3.Connection.search()
        :param kwargs: kwargs to pass to ldap3.Connection.search()
        :return: None
        """
        try:
            self.conn.search(*args, **kwargs)
        except LDAPSessionTerminatedByServerError:
            self.conn.bind()
            self.conn.search(*args, **kwargs)

    def _get_ldap_user(self, uid: Union[str, int]) -> Union[str, int]:
        """
        Get the user listed in LDAP with the given UID

        :param uid: UID to search for
        :return: Username related to UID or UID
        """
        result = uid

        try:
            result = getpwuid(uid).pw_name

        except KeyError:
            self._ldap_query(
                'ou=jasmin,ou=People,o=hpc,dc=rl,dc=ac,dc=uk',
                f'(&(objectClass=posixAccount)(uidNumber={uid}))',
                attributes='uid',
                size_limit=1
            )

            result = self._process_result('uid')

        finally:
            self.users[uid] = result

        return result

    def _get_ldap_group(self, gid: Union[str, int]) -> Union[str, int]:
        """
        Get the group name linked to the given GID
        :param gid: The GID to search for
        :return: Common Name for given GID or GID
        """
        result = gid

        try:
            result = getgrgid(gid).gr_name

        except KeyError:
            self._ldap_query(
                'ou=ceda,ou=Groups,o=hpc,dc=rl,dc=ac,dc=uk',
                f'(&(objectClass=posixGroup)(gidNumber={gid}))',
                attributes='cn',
                size_limit=1
            )

            result = self._process_result('cn')

        finally:
            self.users[gid] = result

        return result

    def get_user(self, uid: Union[str, int]) -> Union[str, int]:
        """
        Either return from the cache, filesystem or search LDAP for the LDAP user
        :param uid: user ID
        :return: Username with UID or UID
        """

        # Try the cache to see if the user ID is stored
        if self.users.get(uid):
            return self.users.get(uid)

        # Try to get the username from the filesystem
        # If the names are not mounted for the uids then
        # query LDAP directly
        return self._get_ldap_user(uid)

    def get_group(self, gid: Union[str, int]) -> Union[str, int]:
        """
        Either return from the cache, filesystem or search LDAP for the LDAP group

        :param gid: group ID
        :return: Groupname with GID or GID
        """

        # Try the cache
        if self.groups.get(gid):
            return self.users.get(gid)

        # Try to get the group name from file system.
        # If names not mounted, query LDAP direct
        return self._get_ldap_group(gid)


def delete_folder(folder):
    try:
        os.rmdir(folder)  # This deletes only empty dirs.
    except OSError as ex:
        if ex.errno == errno.ENOTEMPTY:
            pass


def sanitise_args(config):
    """
    Sanitise command-line configuration.

    :param config: Config dictionary (from docopt)
    :returns: Config dictionary with all keys stripped of '<' '>' and '--'
    """
    sane_conf = {}
    for key, value in six.iteritems(config):
        if value is not None:
            key = key.lstrip("-><").rstrip("><")
            sane_conf[key] = value

    return sane_conf


def read_conf(conf_path):
    """
    Reads configuration file into a dictionary.

    :param conf_path: Path to the JSON configuration file.
    :returns: Dict containing parsed JSON conf.
    """
    try:
        with open(conf_path, "r") as conf:
            return json.load(conf)
    except IOError:
        err_path = os.path.abspath(conf_path)
        sys.stderr.write(  # Continued on next line
            "Can't read configuration file\n%s\n\n" % err_path)
        return {}


def cfg_read(filename):
    """
    Reads configuration file into a dictionary.

    :param filename: Path to the INI configuration file.
    :returns: Dict containing parsed ini conf.
    """
    # Read the config file
    config = ConfigParser()
    config.read(filename)

    # get sections
    sections = config.sections()

    conf = {}
    section_options = {}

    for section in sections:

        options = config.options(section)

        for option in options:

            try:
                value = config.get(section, option)
                parsed_value = value.replace("\"", "")
                section_options[option] = parsed_value

                if section_options[option] == -1:
                    section_options[option] = None
            except:
                section_options[option] = None

        conf[section] = section_options.copy()
        section_options.clear()

    return conf


def get_settings(conf_path, args):
    # Default configuration options
    # These are overridden by the config file and command-line arguments
    defaults = {}

    # conf_file = read_conf(conf_path)
    conf_file = cfg_read(conf_path)

    # print( conf_file )

    # Apply updates to CONFIG dictionary in priority order
    # Configuration priority: CONFIG < CONF_FILE < ARGS
    # (CONFIG being lowest, ARGS being highest)
    defaults.update(conf_file)
    defaults.update(args)

    return defaults


def build_file_list(path):
    """
    :param path : A file path
    :return: List of files contained within the specified directory.
    """
    
    file_list = []
    for root, _, files in os.walk(path):
        for each_file in files:
            if each_file[0] == ".": continue
            if os.path.islink(each_file): continue

            file_list.append(os.path.join(root, each_file))

    return file_list


def write_list_to_file(task_list, filename):
    with open(filename, 'w') as writer:
        writer.writelines(
            map(lambda x: x + "\n", task_list)
        )

    return len(task_list)


def read_file_into_list(filename):
    content = []
    with open(filename) as fd:
        for line in fd:
            content.append(line)
    return content


def save_to_file(filename, data):
    with io.FileIO(filename, "a") as fp:
        fp.write(data)


def find_in_list(list_str, str_item):
    for str in list_str:
        if str_item in str:
            return str

    return None


def get_file_header(filename):
    """
    :param filename : The file to be read.
    :returns: First line of the file.
    """
    with open(filename, 'r') as fd:
        first_line = fd.readline()

    return first_line.replace("\n", "")


def get_bytes_from_file(filename, num_bytes):
    """
    :param filename : The file to be read.
    :param num_bytes : number of bytes to read.
    :returns: The first num_bytes from the file.
    """

    try:
        fd = open(filename, 'r')
        bytes_read = fd.read(num_bytes)
        fd.close()
    except IOError:
        return None

    return bytes_read


def find_dataset(filename, dataset_id):
    """
    :param filename : file containing dataset information.
    :param dataset_id : dataset to be searched.
    :returns: The path of the given dataset id.
    """
    var_dict = {}
    with open(filename) as l_file:
        for line in l_file:
            if not line.startswith("#"):
                name, var = line.partition("=")[::2]
                var_dict[name.strip()] = var.strip()

    if dataset_id == "all":
        return var_dict
    else:
        try:
            return var_dict[dataset_id]
        except KeyError as ex:
            return None


def find_num_lines_in_file(filename):
    """
    :param filename : Name of the file to be read.
    :returns: The number of lines in the given file.
    """
    num_lines = 0

    with open(filename) as infp:
        for line in infp:
            num_lines += 1
    return num_lines


def valid_attr_length(name, value):
    if len(value) < MAX_ATTR_LENGTH \
            and len(name) < MAX_ATTR_LENGTH:
        return True
    return False


def is_valid_phen_attr(attribute):
    if attribute is None:
        return True
    elif re.search(r"\d+-\d+-\d+.*", attribute) is not None:
        return False
    else:
        return True


def is_valid_parameter(name, value):
    valid_parameters = ["standard_name",
                        "long_name",
                        "title",
                        "name",
                        "units",
                        "var_id",
                        "title"
                        ]
    if name in valid_parameters \
            and valid_attr_length(name, value):
        return True
    return False


def is_valid_phenomena(key, value):
    """
    Wrapper to hide test in main function
    """

    if value is None:
        return False

    if not is_valid_parameter(key, value):
        return False

    if not is_valid_phen_attr(value):
        return False

    # Returns true if the tests above pass as true
    return True


def is_date_valid(date_text):
    try:
        datetime.datetime.strptime(date_text, '%Y-%m-%d')
        return True
    except ValueError:
        return False


def get_best_name(phenomena):
    """
    Create a best_name field which takes the best name as defined by the preference order
    :param phenomena: phenomena attributes in form [{"name":"standard_name","value":"time"},{"name":"---","value":"---"},{}...]
    :return: best_name(string)
    """
    preference_order = ["long_name","standard_name","title","name","short_name","var_id"]
    attributes = phenomena["attributes"]

    for name in preference_order:
        best_name = [d['value'] for d in attributes if d['name'] == name]
        if best_name:
            return best_name[0]
    return None


def long_name_is_standard_name(phenomena):
    """
    If the long name and the standard name are the same and the long_name contains _
    return True
    :param phenomena: phenomena attributes in form [{"name":"standard_name","value":"time"},{"name":"---","value":"---"},{}...]
    :return: Boolean
    """
    attributes = phenomena["attributes"]

    long_name = None
    standard_name = None

    for d in attributes:
        if d["name"] == "long_name":
            long_name = d["value"]

        if d["name"] == "standard_name":
            standard_name = d["value"]

    if long_name is None or standard_name is None:
        # If one of them is missing then don't need to do next checks.
        return False

    if long_name.strip() == standard_name.strip() and '_' in long_name:
            return True
    else:
        return False


def build_phenomena(data):
    if not data:
        return (None,)

    phenom_list = []

    name_filter = ["units", "var_id", "standard_name", "long_name"]
    names_list_filter = ["standard_name", "long_name", "title", "name"]

    for phenom in data:
        phen_dict = {}
        names = []
        agg_string = ""
        agg_string_list = []

        best_name = get_best_name(phenom)

        long_name_check = long_name_is_standard_name(phenom)

        for attr in phenom["attributes"]:

            value = attr["value"]
            name = attr["name"]

            if name in name_filter:
                # Remove extra spaces and any quotation marks which will interfere with creating the agg_string
                value = re.sub('  +', ' ', value).replace('"', '')
                phen_dict[name] = value
                agg_string_list.append('"{}":"{}"'.format(name, value))

            if name in names_list_filter and value not in names:

                # if long_name contains "_" and is the same as standard name, include a version in the names
                # list with the "_" replaced by " "
                if name == "long_name" and long_name_check:
                    value = value.replace("_"," ")

                names.append('"{}"'.format(value))

            phen_dict[name] = value

        if names:
            names.sort()
            phen_dict["names"] = names
            agg_string_list.append('"names":{}'.format(';'.join(names)))

        if agg_string_list:
            agg_string_list.sort()
            agg_string = ','.join(agg_string_list)

        if best_name:
            phen_dict["best_name"] = best_name

        if phen_dict:
            phen_dict["agg_string"] = agg_string
            phenom_list.append(phen_dict)

    return (phenom_list,)


def date2iso(date):
    date = parser.parse(date)
    iso_date = date.isoformat()

    return iso_date


def calculate_md5(filename):
    hash_md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


# encoding: utf-8
__author__ = 'Daniel Westwood'
__date__ = '05 Nov 2024'
__copyright__ = 'Copyright 2024 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'daniel.westwood@stfc.ac.uk'

# Opens the file provided in the datasets folder.
# Iterate over all provided dataset paths.

# facet_scanner.get_handler('filepath')
# facets = handler.get_facets('filepath')

import logging
import hashlib
import argparse
import os
from datetime import datetime
import glob

import asyncio

from cci_facet_scanner.core.facet_scanner import FacetScanner
from elasticsearch import Elasticsearch

from cci_os_worker.filehandlers.util import LDAPIdentifier
from cci_os_worker.filehandlers import NetCdfFile, GenericFile
from cci_os_worker import logstream

from .path_tools import PathTools
from .directory import check_timeout
from .utils import load_config, UpdateHandler, set_verbose
from .errors import HandlerError, DocMetadataError

from slack_sdk import WebClient

logger = logging.getLogger(__name__)
logger.addHandler(logstream)
logger.propagate = False

class FacetUpdateHandler(UpdateHandler):

    def __init__(self, conf: dict, dryrun: bool = False, test: bool = False):
        """
        Initialise this class with the correct connections to 
        establish an elasticsearch client.
        """
        logger.info('Loading Facet Updater')

        super().__init__(conf, dryrun=dryrun, test=test)

        self.facet_scanner = FacetScanner()

        ldap_hosts = self._conf['ldap_configuration']['hosts']
        self.ldap_interface = LDAPIdentifier(server=ldap_hosts, auto_bind=True)

        self._spot_file = conf.get('spot_file',None)
        self.pt = PathTools(spot_file=self._spot_file)

    def _get_project_info(self, path):
        """
        Get project info for a specific path
        """

        extension = os.path.splitext(path.split('/')[-1])[-1]
        extension = extension.lower()

        if extension == '.nc':
            handler = NetCdfFile
        else:
            handler = GenericFile

        calculate_md5 = self._conf.get('calculate_md5',False)

        if handler is None:
            raise HandlerError(filename=path)

        handler_instance = handler(path, 3, calculate_md5=calculate_md5)

        # FutureDetail: Remove manifest from 'doc' if unneeded (no indexing required.)
        doc, phenomena, spatial = handler_instance.get_metadata()

        if doc is None:
            raise DocMetadataError(filename=path)
        if len(doc) > 1:
            doc = doc[0]

        if phenomena:
            doc['info']['phenomena'] = phenomena
        if spatial:
            doc['info']['spatial'] = spatial

        spot = self.pt.spots.get_spot(path)

        if spot is not None:
            doc['info']['spot_name'] = spot

        # Replace the UID and GID with name and group
        uid = doc['info']['user']
        gid = doc['info']['group']

        doc['info']['user'] = self.ldap_interface.get_user(uid)
        doc['info']['group'] = self.ldap_interface.get_group(gid)

        return doc['info']

    def _single_process_file(self, filepath: str, index: int = None, total: int = None):
        """
        Perform facet scanning for a specific filepath
        """

        logger.info('--------------------------------')
        if index is None:
            logger.info(f'Processing {filepath.split("/")[-1]}')
        else:
            logger.info(f'Processing {filepath.split("/")[-1]} ({index}/{total})')

        # Get the handler for this filepath
        handler = self.facet_scanner.get_handler(filepath, json_files=None)

        # Extract the facets
        facets = handler.get_facets(filepath)

        # Build the project dictionary using the handlers project name attr
        project = {
            'info': self._get_project_info(filepath),
            'projects': {
                handler.project_name: facets
            }
        }

        if self._test:
            index = self._conf['facet_files_test_index']['name']
        else:
            index = self._conf['facet_files_index']['name']

        id = hashlib.sha1(filepath.encode(errors="ignore")).hexdigest()

        # Send facets to elasticsearch
        if not self._dryrun:
            self.es.update(
                index=index,
                id=id,
                body={'doc': project, 'doc_as_upsert': True}
            )
        else:
            logger.info(f'DRYRUN: Skipped updating for {filepath.split("/")[-1]}')

            self._local_cache(
                filename=f'cache/{filepath.split("/")[-1]}-cache.json',
                contents=project,
            )

def _get_command_line_args():
    """
    Get the command line arguments for the facet scan
    """
    parser = argparse.ArgumentParser(description='Entrypoint for the CCI OS Worker on the CMD Line')
    parser.add_argument('datafile_path', type=str, help='Path to the "datasets.txt" file')
    parser.add_argument('conf', type=str, help='Path to Yaml config file for Elasticsearch')

    parser.add_argument('-d','--dryrun', dest='dryrun', action='store_true', help='Perform in dryrun mode')
    parser.add_argument('-t','--test', dest='test', action='store_true', help='Perform in test/staging mode')
    parser.add_argument('-p','--prefix', dest='prefix', default='', help='Prefix to apply to all filenames')
    parser.add_argument('-v','--verbose', action='count', default=2, help='Set level of verbosity for logs')
    parser.add_argument('-f','--file-count', dest='file_count', type=int, help='Add limit to number of files to process.')
    parser.add_argument('-o','--output', dest='output', default=None, help='Send fail list to an output file')

    args = parser.parse_args()

    return {
        'datafile_path': args.datafile_path,
        'conf': args.conf,
        'dryrun': args.dryrun,
        'test': args.test,
        'prefix': args.prefix,
        'verbose': args.verbose-1,
        'file_count': args.file_count,
        'output': args.output
    }

def get_startup_slack(timestamp: str, file_count: int, is_sample: bool):
    """
    Get the startup message for the job.
    """
    job_summary = ['----------------------------------------------']
    job_summary.append(f'Starting CCI OS Worker Job ({timestamp})')
    job_summary.append('')
    job_summary.append('Json files:')
    jsons_new = os.environ.get("JSON_TAGGER_NEW")
    jsons = glob.glob(f'{jsons_new}/*.json')
    for js in jsons:
        job_summary.append(
            f' - {js.split("/")[-1]}'
        )
    job_summary.append('')

    ds_count = f'Datasets identified: {file_count}'
    if is_sample:
        ds_count += ' (sample run)'
    job_summary.append(ds_count)
    job_summary.append('----------------------------------------------')

    return '\n'.join(job_summary)

def get_completion_slack(timestamp: str, fail_list: list, num_jobs: int):
    """
    Get the completion message for the job"""
    job_complete = ['----------------------------------------------']
    job_complete.append(f'CCI OS Worker Job Complete! ({timestamp})')
    job_complete.append(f' - Datasets tagged: {num_jobs}')
    if len(fail_list) == 0: 
        job_complete.append(f' - No Failures detected! :)')
    else:
        job_complete.append(f' - {len(fail_list)} dataset(s) failed to scan :( ')
    job_complete.append('----------------------------------------------')
    return '\n'.join(job_complete)


def main(args: dict = None):
    if args is None:
        args = _get_command_line_args()
    if isinstance(args['conf'], str):
        conf = load_config(args['conf'])
    
    slack_cfg = conf.get('slack_cfg',None)
    slack_client = None

    timestamp = datetime.now().strftime("%H:%M:%S %d/%m/%y")

    is_sample = True
    file_count = conf.get('file_limit',None)
    if file_count is None:
        is_sample = False

        with open(args['datafile_path']) as f:
            file_count = len(f.readlines())

    if slack_cfg is not None:

        token = slack_cfg['token']
        channel = slack_cfg['channel']
        slack_client = WebClient(token=token)
        slack_client.chat_postMessage(
            channel=channel, 
            text=get_startup_slack(timestamp, file_count, is_sample),
            username=f'CCI Tag Bot - Startup'
        )
        logger.info('Message posted!')

    if conf is None:
        logger.error('Config file could not be loaded')
        return
    if not os.path.isfile(args['datafile_path']):
        logger.error(f'Inaccessible Datafile - {args["datafile_path"]}')
        return
    
    if check_timeout():
        logger.error('Check-timeout failed')
        return
    
    file_limit = conf.get('file_limit', None) or args.get('file_limit', None)

    set_verbose(args['verbose'])

    fs = FacetUpdateHandler(conf, dryrun=args['dryrun'], test=args['test'])
    fail_list = fs.process_deposits(args['datafile_path'], args['prefix'], file_limit=file_limit)

    logger.info('Failed items:')
    for f in fail_list:
        logger.info(f)

    if args['output'] is not None and fail_list != []:
        with open(args['output'],'w') as f:
            f.write('\n'.join(fail_list))

    if slack_client is not None:

        slack_client.chat_postMessage(
            channel=channel, 
            text=get_completion_slack(timestamp, fail_list, file_count),
            username=f'CCI Tag Bot - Complete'
        )

if __name__ == '__main__':
    main()